# -*- coding: utf-8 -*-
"""Tools : database of spectra, line survey, interface with Cantera.
"""


from .database import SpecDatabase, load_spec, plot_spec, save
from .gascomp import get_eq_mole_fraction
from .slit import (
    convolve_with_slit,
    crop_slit,
    get_effective_FWHM,
    get_FWHM,
    plot_slit,
    recenter_slit,
)

__all__ = [
    "SpecDatabase",
    "load_spec",
    "plot_spec",
    "save",
    "get_eq_mole_fraction",
    "plot_slit",
    "get_effective_FWHM",
    "get_FWHM",
]



# -*- coding: utf-8 -*-
"""
Functions and constants used in :class:`~radis.spectrum.spectrum.Spectrum`
object

-------------------------------------------------------------------------------


"""

import subprocess

from radis.misc.basics import partition

# %% Definitions

# Waverange, wavespaces
# ... aliases:
WAVENUM_UNITS = ["cm", "cm-1"]
WAVELEN_UNITS = ["nm", "nm_air"]
WAVELENVAC_UNITS = ["nm_vac", "nm_vacuum"]
# ... internal names
WAVESPACE = ["nm", "nm_vac", "cm-1"]
""" list: wavespace:
- ``'nm'``: wavelength (in air)
- ``'nm_vac'``: wavelength (in vacuum)
- ``'cm-1'``: wavenumber"""

# Spectral quantities
CONVOLUTED_QUANTITIES = ["radiance", "transmittance", "emissivity"]
"""list: name of spectral quantities after convolution with slit function"""
NON_CONVOLUTED_QUANTITIES = [
    "radiance_noslit",
    "transmittance_noslit",
    "emisscoeff",
    "emisscoeff_continuum",
    "absorbance",
    "abscoeff",  # opacity
    "abscoeff_continuum",
    "emissivity_noslit",
    "xsection",  # cross-sections
]
"""list: name of spectral quantities not convolved with slit function

See :ref:`the description of spectral arrays <label_spectral_arrays>`"""
SPECTRAL_QUANTITIES = CONVOLUTED_QUANTITIES + NON_CONVOLUTED_QUANTITIES
"""list: all spectral quantities defined in a :class:`~radis.spectrum.spectrum.Spectrum`
object.

See :ref:`the description of spectral arrays <label_spectral_arrays>`"""

# note: it is hardcoded (and needed) that quantities that are convoluted are
# generated from a non convoluted quantity with the same name + _noslit
for _q in CONVOLUTED_QUANTITIES:
    assert _q + "_noslit" in NON_CONVOLUTED_QUANTITIES

# %% Special names when printing conditions

# These parameters are shown below "Physical Conditions" rather than "Computation
# Parameters
PHYSICAL_PARAMS = [
    "species",
    "wavenum_max",
    "wavenum_min",
    "mole_fraction",
    "isotope",
    "state",
    "path_length",
    "medium",
    "self_absorption",
    "slit_function_base",
    "pressure",
    "wavelength_min",
    "wavelength_max",
    "Telec",
    "Tvib",
    "Trot",
    "Tgas",
    "vib_distribution",
    "rot_distribution",
    "overpopulation",
    "thermal_equilibrium",
]
"""list: physical conditions under which the Spectrum was calculated/measured.
When printing an object, these parameters are shown below "Physical Conditions"
If a parameter is not in this list, it is either in "Computation Parameters"
(non-physical parameters that can have an influence on the Spectrum, e.g, cutoffs
and thresholds) or in "Informative Params" (descriptive parameters that have absolutely no
impact on the spectrum, e.g, number of lines calculated or calculation time)
"""

# Informative parameters
# ... Parameters that should be saved in the Spectrum objects, but ignored when
# ... comparing two spectra.
# ... should be written here only these parameters that cannot affect the
# ... physical result. In particular, all parameters relative to performance
# ... should be added here.
INFORMATIVE_PARAMS = [
    "db_use_cached",
    "lvl_use_cached",
    "chunksize",
    "calculation_time",
    "total_lines",
    "lines_calculated",
    "lines_cutoff",
    "lines_in_continuum",
    "Nprocs",
    "warning_broadening_threshold",
    "warning_linestrength_cutoff",
    "load_energies",
    "export_lines",
    "export_populations",
    # parameters added to reduce default verbose :
    "levelsfmt",
    "wavenum_max_calc",
    "wavenum_min_calc",
    "dxG",  # TODO : keep it a Computational parameter, but of verbose-level 2 so it's not shown by default
    "dxL",
    "export_rovib_fraction",
    "parfuncpath",
]
""" list: Informative parameters. Parameters that should be saved in the Spectrum
objects, but ignored when comparing two spectra. Should be written here only
these parameters that cannot affect the physical result. In particular, all
parameters relative to performance should be added here.

Notes
-----

units for these parameters are stored in Spectrum.cond_units and are defined
by the generating class (ex: SpectrumFactory)"""

CONFIG_PARAMS = [
    "GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD",
    "GRIDPOINTS_PER_LINEWIDTH_ERROR_THRESHOLD",
    "SPARSE_WAVERANGE",
    "DEFAULT_DOWNLOAD_PATH",
]
""" list: these parameters are read from radis.config and stored in the Spectrum
objects. Should be added here only the parameters that may have an impact on the computation,
for instance the one defining the 'auto' thresholds
"""

# %% Util functions


def cast_waveunit(unit, force_match=True):
    """Standardize unit formats, return either "nm", "nm_vac" or "cm-1"."""
    if unit in WAVELEN_UNITS:
        return "nm"
    if unit in WAVELENVAC_UNITS:
        return "nm_vac"
    elif unit in WAVENUM_UNITS:
        return "cm-1"
    elif force_match:
        raise ValueError(
            "Unknown wavespace unit: {0}. Should be one of {1}".format(
                unit, WAVELEN_UNITS + WAVELENVAC_UNITS + WAVENUM_UNITS
            )
        )
    else:
        return unit  # dont convert


def make_up(label):
    """Cosmetic changes on label, before plot.

    Parameters
    ----------

    label: str
    """

    # Hardfix for missing characters
    SUPERSCRIPT_MINUS = "⁻"
    from matplotlib.font_manager import FontProperties, findfont

    if "arial" in findfont(FontProperties(family=["sans-serif"])):
        SUPERSCRIPT_MINUS = "$^{-}$"

    # Improve units
    label = label.replace(r"cm-1", r"cm" + f"{SUPERSCRIPT_MINUS}" + r"¹")
    label = label.replace(r"m^-1", r"m" + f"{SUPERSCRIPT_MINUS}" + r"¹")
    label = label.replace(r"m2", r"m²")
    label = label.replace(r"m3", r"m³")
    label = label.replace(r"I/I0", r"I/I₀")  # transmittance unit

    # Improve text
    if not "_noslit" in label:
        # small hack: capitalize, to make a difference with non slit value
        label = label.replace(
            "radiance", "Radiance"
        )  # make a small difference between no_slit and slit while plotting
        label = label.replace(
            "transmittance", "Transmittance"
        )  # make a small difference between no_slit and slit while plotting

    label = label.replace("xsection", "cross-section")

    # ... Remove _noslit
    label = label.replace("_noslit", "")

    return label


def make_up_unit(Iunit, var):
    """Additional cosmetic changes for units on label, before plot.

    Parameters
    ----------

    Iunit: str
        input unit

    var: str
        spectral variable. Ex: ``transmittance``
    """
    Iunit = Iunit.replace(r"um", r"µm")
    Iunit = Iunit.replace(r"cm-1", r"cm⁻¹")  # welcome to unicode ! ;)
    Iunit = Iunit.replace(r"m2", r"m²")

    if Iunit == "":
        # give more explicit unit for the user:
        if var in ["transmittance", "transmittance_noslit"]:
            Iunit = r"I/I0"
        elif var == "absorbance":
            Iunit = r"-ln(I/I0)"
        elif var in ["emissivity_no_slit", "emissivity"]:
            Iunit = r"ε"
        elif var in ["radiance", "radiance_noslit"]:
            Iunit = r"norm"

    return Iunit


def format_xlabel(wunit, plot_medium):
    """Used by :py:meth:`radis.spectrum.spectrum.Spectrum.plot` and
    :py:func:`radis.spectrum.compare.plot_diff`

    Parameters
    ----------

    wunit: ``'default'``, ``'nm'``, ``'cm-1'``, ``'nm_vac'``,
        wavelength air, wavenumber, or wavelength vacuum. If ``'default'``,
        Spectrum :py:meth:`~radis.spectrum.spectrum.Spectrum.get_waveunit` is used.

    plot_medium: bool, ``'vacuum_only'``
        if ``True`` and ``wunit`` are wavelengths, plot the propagation medium
        in the xaxis label (``[air]`` or ``[vacuum]``). If ``'vacuum_only'``,
        plot only if ``wunit=='nm_vac'``. Default ``'vacuum_only'``
        (prevents from inadvertently plotting spectra with different propagation
        medium on the same graph).

    """
    if wunit == "cm-1":
        xlabel = "Wavenumber (cm-1)"
    elif wunit == "nm":
        if plot_medium and plot_medium != "vacuum_only":
            xlabel = "Wavelength [air] (nm)"
        else:
            xlabel = "Wavelength (nm)"
    elif wunit == "nm_vac":
        if plot_medium and plot_medium != "vacuum_only":
            xlabel = "Wavelength [vacuum] (nm)"
        else:
            xlabel = "Wavelength (nm)"
    else:
        raise ValueError(wunit)

    return make_up(xlabel)


def print_conditions(
    conditions,
    units,
    phys_param_list=PHYSICAL_PARAMS,
    info_param_list=INFORMATIVE_PARAMS,
    config_param_list=CONFIG_PARAMS,
    verbose=2,
):
    """Print all Spectrum calculation parameters.

    Parameters
    ----------
    phys_param_list: list
        These parameters are shown below "Physical Conditions" rather than "Computation
        Parameters. See :data:`~radis.spectrum.utils.PHYSICAL_PARAMS` for more
        information.
    info_param_list: list
        These parameters are shown below "Information" rather than "Computation
        Parameters. See :data:`~radis.spectrum.utils.INFORMATIVE_PARAMS` for more
        information.
    config_param_list: list
        These parameters are read from radis.config file. See
        :data:`~radis.spectrum.utils.CONFIG_PARAMS` for more
        information.
    verbose: int
        if ``1`` or ``True``, only physical and computational parameters are shown. If ``2``,
        all parameters (including config & informative) are shown. Default ``2``.

    See Also
    --------

    :data:`~radis.spectrum.utils.PHYSICAL_PARAMS`, :data:`~radis.spectrum.utils.INFORMATIVE_PARAMS`
    """

    def align(a, space=20):
        r"""fix alignment."""
        return a + " " * max(1, (space - len(str(a))))

    def print_param(k):
        r"""Special formatting for nicely printing conditions."""
        v_k = conditions[k]
        # Add extra arguments based on arbitrary conditions
        args = []
        if k in units:
            args.append(units[k])
        # ... fill here for other args

        # Special formatting
        try:
            if k in [
                "wavenum_max_calc",
                "wavenum_min_calc",
                "wavelength_max",
                "wavelength_min",
                "wavenum_max",
                "wavenum_min",
            ]:
                v_k_str = "{0:.4f}".format(v_k)
            elif k in ["lines_calculated", "lines_in_continuum", "lines_cutoff"]:
                # Add comma separator for thousands
                v_k_str = "{0:,d}".format(v_k)
            else:
                # Default to printing str
                v_k_str = "{0}".format(v_k)
        except ValueError:
            # Default to printing str
            v_k_str = "{0}".format(v_k)

        # Crop
        if len(v_k_str) > 102:  # cut if too long
            v_k_str = v_k_str[:100] + "..."
        print(" " * 2, align(k), v_k_str, *args)

    phys_param, non_phys_param = partition(lambda x: x in phys_param_list, conditions)

    info_param, non_phys_param = partition(
        lambda x: x in info_param_list, non_phys_param
    )

    config_param, non_phys_param = partition(
        lambda x: x in config_param_list, non_phys_param
    )

    # TODO : add verbose level so that Trot, Tvib are not shown if equilibrium
    # and verbose<2 . Same for rot_distribution; vib_distribution

    print("Physical Conditions")
    print("-" * 40)
    for k in sorted(phys_param):
        print_param(k)

    print("Computation Parameters")
    print("-" * 40)
    for k in sorted(non_phys_param):
        print_param(k)

    if verbose >= 2:
        print("Config parameters")
        print("-" * 40)
        for k in sorted(config_param):
            print_param(k)

        if len(info_param) > 0:
            print("Information")
            print("-" * 40)
            for k in sorted(info_param):
                print_param(k)

    print("-" * 40)

    if verbose >= 2:

        # print gas_inp (information on each gas slab) if exists (specifically
        # for Specair output)
        if "gas_inp" in conditions:
            try:
                for slab in conditions["gas_inp"]:
                    print("Slab", slab)
                    slab.print_conditions()
            except:
                pass

    return None


# %% Plot helper


def split_and_plot_by_parts(w, I, *args, **kwargs):
    from warnings import warn

    from radis.misc.plot import split_and_plot_by_parts

    warn(
        DeprecationWarning(
            "split_and_plot_by_parts() moved in radis.misc.plot in 0.9.30"
        )
    )
    return split_and_plot_by_parts(w, I, *args, **kwargs)


# %% Profiler

#%%
def dict_to_tree(pro, name):
    """
    Parameters
    ----------
    pro: dict
        of the form::

            {"value":float,    # keyword 'value' is expected
             "some_key":float
             "some_key2":float,
             "some_key3":dict  # nested dict of the same form
             "some_key4":dict}
    name: str

    Returns
    -------
    dict: of the form ::
        Tree =
        {"name": str,
         "value":float,
         "children":[list of Tree]}

    See Also
    --------
    Used in :py:func:`~radis.spectrum.utils.print_perf_profile` and
    :py:func:`~radis.spectrum.utils.generate_perf_profile`

    """
    if isinstance(pro, dict):
        new_dict = {"value": pro["value"], "name": name}
        children = {k: v for k, v in pro.items() if k != "value"}
        if len(children) > 0:
            new_dict["children"] = [
                dict_to_tree(v, name=k) for k, v in children.items()
            ]
        return new_dict
    else:
        return {"value": pro, "name": name}


#%%


def print_perf_profile(
    profiler,
    total_time=None,
    number_format="{:.3f}",
    precision=16,
    first_line="profiler :",
):
    r"""Prints Profiler output dictionary in a structured manner.

    Parameters
    ----------
    profiler: dict
        of the form::
            {"value":float,    # keyword 'value' is expected
             "some_key":float
             "some_key2":float,
             "some_key3":dict  # nested dict of the same form
             "some_key4":dict}

    Other Parameters
    ----------------
    total_time: float
        total calculation time. If ``None``, take the key ``"value"`` of ``profiler``
    precision: int, optional
        total number of blocks. Default 16.

    Example
    -------
    ::

        Spectrum.print_perf_profile()

        # output >>
            spectrum_calculation      0.189s ████████████████
                check_line_databank              0.000s
                check_non_eq_param               0.042s ███
                fetch_energy_5                   0.015s █
                calc_weight_trans                0.008s
                reinitialize                     0.002s
                    copy_database                    0.000s
                    memory_usage_warning             0.002s
                    reset_population                 0.000s
                calc_noneq_population            0.041s ███
                    part_function                    0.035s ██
                    population                       0.006s
                scaled_non_eq_linestrength       0.005s
                    map_part_func                    0.001s
                    corrected_population_se          0.003s
                calc_emission_integral           0.006s
                applied_linestrength_cutoff      0.002s
                calc_lineshift                   0.001s
                calc_hwhm                        0.007s
                generate_wavenumber_arrays       0.001s
                calc_line_broadening             0.074s ██████
                    precompute_LDM_lineshapes        0.012s
                    LDM_Initialized_vectors          0.000s
                    LDM_closest_matching_line        0.001s
                    LDM_Distribute_lines             0.001s
                    LDM_convolve                     0.060s █████
                    others                           0.001s
                calc_other_spectral_quan         0.003s
                generate_spectrum_obj            0.000s
                others                           -0.016s

    See Also
    --------
    :py:meth:`~radis.spectrum.spectrum.Spectrum.print_perf_profile`,
    :py:meth:`~radis.lbl.factory.SpectrumFactory.print_perf_profile`
    """
    PRECISION = precision  #  number of blocks █
    TAB = 4  # number of spaces for indentation

    if total_time is None:
        total_time = profiler["value"]

    def scale(time):
        # return '|'*int(time/total_time*10)
        return "\u2588" * int(time / total_time * PRECISION)
        # '\u2588'  is  █

    def walk_print_tree(prof, name, level, write_number_column):
        if "value" in prof:
            text = " " * TAB * level + name
            fill_spaces = " " * (write_number_column - len(text))
            print(
                text,
                fill_spaces,
                "" + number_format.format(prof["value"]) + "s",
                scale(prof["value"]),
            )
        max_name_length = max(len(k) for k in prof.keys() if k != "value")
        write_number_column = max(
            write_number_column, len(" " * TAB * (level + 1)) + max_name_length
        )
        write_number_column += TAB

        total_time = 0
        for k, v in prof.items():
            if k == "value":
                pass
            elif isinstance(v, dict):
                total_time += walk_print_tree(
                    v,
                    name=k,
                    level=level + 1,
                    write_number_column=write_number_column,
                )
            elif isinstance(v, float):
                text = " " * TAB * (level + 1) + k
                fill_spaces = " " * (write_number_column - len(text))
                print(text, fill_spaces, "" + number_format.format(v) + "s", scale(v))
                total_time += v
            else:
                raise ValueError(type(v))

        # print missing time / self-time
        if "value" in prof:
            missing_time = prof["value"] - total_time
            if float(number_format.format(missing_time)) != 0:
                # we dont add 0 numbers
                text = " " * TAB * (level + 1) + "others"
                fill_spaces = " " * (write_number_column - len(text))
                print(
                    text,
                    fill_spaces,
                    "" + number_format.format(missing_time) + "s",
                    scale(missing_time),
                )

        return total_time

    print(first_line)
    walk_print_tree(profiler, name="", level=0, write_number_column=0)


def generate_perf_profile(profiler):
    """Visual/interactive performance profile

    Requires ``tuna`` to be installed.

    See typical output in https://github.com/radis/radis/pull/325

    .. image:: https://user-images.githubusercontent.com/16088743/128018032-6049be72-1881-46ac-9d7c-1ed89f9c4f42.png
        :alt: https://user-images.githubusercontent.com/16088743/128018032-6049be72-1881-46ac-9d7c-1ed89f9c4f42.png
        :target: https://user-images.githubusercontent.com/16088743/128018032-6049be72-1881-46ac-9d7c-1ed89f9c4f42.png


    .. note::
        You can also profile with `tuna` directly::

            python -m cProfile -o program.prof your_radis_script.py
            tuna your_radis_script.py

    Parameters
    ----------
    profiler: dict
        of the form::
            {"value":float,    # keyword 'value' is expected
             "some_key":float
             "some_key2":float,
             "some_key3":dict  # nested dict of the same form
             "some_key4":dict}

    See Also
    --------
    :py:meth:`~radis.spectrum.generate_perf_profile`
    """
    from pstats import Stats

    st = Stats()

    """
    From Lib/profile.py:Profile

        Timing data for each function is stored as a 5-tuple in the dictionary
        self.timings[].  The index is always the name stored in self.cur[-3].
        The following are the definitions of the members:

        [0] = The number of times this function was called, not counting direct
              or indirect recursion,
        [1] = Number of times this function appears on the stack, minus one
        [2] = Total time spent internal to this function
        [3] = Cumulative time that this function was present on the stack.  In
              non-recursive functions, this is the total execution time from start
              to finish of each invocation of a function, including time spent in
              all subfunctions.
        [4] = A dictionary indicating for each function name, the number of times
              it was called by us.
    """

    def parse_profiler(tree, parent=None, parent_time=None):
        r"""

        Parameters
        ----------
        tree : dict
            ::
            {"name": str,
             "value":float,
             "children":[list of Tree]}
        parent : tuple (str, int, str)
            filename(str), line_number(int), parent function name(str)
        parent_time : int, int,
            ``cc, nc, tt, ct`` of parent function.
            - The number of times this function was called, not counting direct
              or indirect recursion,
            - Number of times this function appears on the stack, minus one
            - Total time spent internal to this function
            - Cumulative time that this function was present on the stack.

        """
        func_name = (
            "",
            1,
            tree["name"],
        )  # normally : filename(str), line_number(int), function name (str)

        # Parse children
        if "children" in tree:
            children_cumtime = sum([child["value"] for child in tree["children"]])
            for child in tree["children"]:
                # child_name = "", 1, child["name"]
                child_time = child["value"]
                parse_profiler(
                    child, parent=func_name, parent_time=(1, 0, child_time, child_time)
                )

            if parent:
                parent_time = (
                    parent_time[0],
                    parent_time[1],
                    parent_time[2] - children_cumtime,
                    parent_time[3],
                )
            tt = tree["value"] - children_cumtime
            ct = tree["value"]

        else:
            tt = tree["value"]
            ct = tree["value"]

        # Store value
        st.stats[func_name] = (1, 0, tt, ct, {parent: parent_time} if parent else {})

    perf_tree = dict_to_tree(profiler, name="calculation_time")

    parse_profiler(perf_tree)
    st.dump_stats("spectrum.prof")

    return subprocess.Popen(["tuna", "spectrum.prof"])


if __name__ == "__main__":
    from radis.test.spectrum.test_utils import test_perf_profile

    test_perf_profile()



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 31 18:00:54 2018

@author: erwan
"""

import pytest

from radis.db.molecules import Molecules, getMolecule
from radis.db.utils import (
    _get_rovib_coefficients,
    get_default_jsonfile,
    get_herzberg_coefficients,
)


@pytest.mark.fast  # this is a fast test. Run fast tests only with 'pytest -m fast'
def test_herzberg_coefficients_CO2_json(verbose=True, *args, **kwargs):
    """Check that default coefficients are found correctly"""
    # @dev: update this test if we ever are to change the default Herzberg
    # coeffs in CO2/molecules_data.json

    assert set(get_herzberg_coefficients("CO2", 1, "X1SIGu+")) == set(
        ["we1", "we2", "we3", "wexe1", "wexe2", "wexe3", "Be", "De", "He"]
    )

    if verbose:
        print(
            (
                "Herzberg coefficients defined for CO2(X)(iso=1):",
                get_herzberg_coefficients("CO2", 1, "X1SIGu+"),
            )
        )


@pytest.mark.fast  # this is a fast test. Run fast tests only with 'pytest -m fast'
def test_getMolecule(verbose=True, *args, **kwargs):

    # Input
    molec = "CO"
    iso = 1
    state = "X"

    # get predefined
    S = getMolecule(molec, iso, state)

    assert S.get_fullname() == "CO(X1Σ+)(iso1)"

    #    # Print CO Dunham development calculation (can be accessed from console)
    #    if verbose:
    #        print('Dunham development')
    #        print('------------\n')
    #    S.get_ref('Erovib')

    return True


# Test defined in neq 0.9.24. Deactivated in the move to RADIS 1.0
# @pytest.mark.fast       # this is a fast test. Run fast tests only with 'pytest -m fast'
# def test_CO_energies_Herzberg_vs_Dunham(verbose=True, *args, **kwargs):
#
#    from radis.db.molecules import CO_X_iso1_Herzberg, CO_X_iso1
#
#    for (v, J) in [(0, 0), (1, 10), (10, 10)]:
#        if verbose:
#            print('v={0}, J={1}'.format(v,J))
#            print('CO_X_iso1_Herzberg: {0:.3f}cm-1'.format(CO_X_iso1_Herzberg.Erovib(v,J)))
#            print('CO_X_iso1_from_json: {0:.3f}cm-1'.format(CO_X_iso1.Erovib(v,J)))
#        assert np.isclose(CO_X_iso1_Herzberg.Erovib(v,J), CO_X_iso1.Erovib(v,J),
#                          rtol=1e-4)


def test_ZPE(verbose=True, *args, **kwargs):
    """Test that Zero-Point-Energy stored in the default RADIS database
    corresponds to the Energy calculated for vi=0, J=0 for all molecules and isotopes
    and electronic states in the RADIS :py:data:`~radis.db.molecules.Molecules`
    list.
    """
    for M, isotopes in Molecules.items():
        if M != "CO2":
            continue
        for iso, elecstates in isotopes.items():
            for state, ElecState in elecstates.items():
                jsonfile = get_default_jsonfile(M)
                coeffs = _get_rovib_coefficients(
                    M,
                    iso,
                    ElecState.get_statename_utf(),
                    jsonfile=jsonfile,
                    remove_trailing_cm1=True,
                )
                if "ZPE" in coeffs:
                    # Calculate ZPE by setting all parameters to 0. Note that
                    # the number of parameters in ElecState.Erovib depends on
                    # the molecule. @dev: we use inspect to get the number of
                    # parameters
                    from radis.misc.utils import getarglist

                    params = [
                        k
                        for k in getarglist(ElecState._Erovib)
                        if k not in ["remove_ZPE", "coeff_dict"]
                    ]
                    # Calculate ZPE:
                    params = [0] * len(params)
                    assert ElecState.Erovib(*params, remove_ZPE=False) == coeffs["ZPE"]
                    if verbose:
                        print(
                            "ZPE calculated for {0} matches value in database ({1:.2f} cm-1)".format(
                                ElecState.get_fullname(), coeffs["ZPE"]
                            )
                        )
                else:
                    if verbose:
                        print(f"No ZPE defined for {ElecState.get_fullname()}")


def _run_testcases(verbose=True, *args, **kwargs):
    #
    test_herzberg_coefficients_CO2_json(verbose=verbose, *args, **kwargs)
    test_getMolecule(verbose=verbose, *args, **kwargs)
    test_ZPE(verbose=verbose, *args, **kwargs)
    #    test_CO_energies_Herzberg_vs_Dunham(verbose=verbose, *args, **kwargs)

    return True


if __name__ == "__main__":

    _run_testcases()



# -*- coding: utf-8 -*-
"""
Created on Sun May 22 18:11:23 2022

@author: erwan
"""

import pathlib

import numpy as np

from radis.api.exomolapi import (
    MdbExomol,
    get_exomol_database_list,
    get_exomol_full_isotope_name,
)
from radis.db.classes import get_molecule_identifier


def fetch_exomol(
    molecule,
    database=None,
    local_databases=None,
    databank_name="EXOMOL-{molecule}",
    isotope="1",
    load_wavenum_min=None,
    load_wavenum_max=None,
    columns=None,
    cache=True,
    verbose=True,
    clean_cache_files=True,
    return_local_path=False,
    return_partition_function=False,
    engine="default",
    output="pandas",
    skip_optional_data=True,
):
    """Stream ExoMol file from EXOMOL website. Unzip and build a HDF5 file directly.

    Returns a Pandas DataFrame containing all lines.

    Parameters
    ----------
    molecule: ``str``
        ExoMol molecule
    database: ``str``
        database name. Ex: ``POKAZATEL`` or ``BT2`` for ``H2O``. See
        :py:data:`~radis.api.exomolapi.KNOWN_EXOMOL_DATABASE_NAMES`. If ``None`` and
        there is only one database available, use it.
    local_databases: ``str``
        where to create the RADIS HDF5 files. Default ``"~/.radisdb/exomol"``.
        Can be changed in ``radis.config["DEFAULT_DOWNLOAD_PATH"]`` or in ~/radis.json config file
    databank_name: ``str``
        name of the databank in RADIS :ref:`Configuration file <label_lbl_config_file>`
        Default ``"EXOMOL-{molecule}"``
    isotope: ``str`` or ``int``
        load only certain isotopes, sorted by terrestrial abundances : ``'1'``, ``'2'``,
        etc. Default ``1``.

        .. note::

            In RADIS, isotope abundance is included in the line intensity
            calculation. However, the terrestrial abundances used may not be
            relevant to non-terrestrial applications.
            By default, the abundance is given reading HITRAN data. If the molecule
            does not exist in the HITRAN database, the abundance is read from
            the ``radis/radis_default.json`` configuration file, which can be
            modified by editing ``radis.config`` after import or directly
            by editing the user ``~/radis.json`` user configuration file
            (overwrites ``radis_default.json``). In the ``radis/radis_default.json``
            file, values were calculated with a simple model based on the
            terrestrial isotopic abundance of each element.

    load_wavenum_min, load_wavenum_max: float (cm-1)
        load only specific wavenumbers.
    columns: list of str
        list of columns to load. If ``None``, returns all columns in the file.

    Other Parameters
    ----------------
    cache: bool, or ``'regen'`` or ``'force'``
        if ``True``, use existing HDF5 file. If ``False`` or ``'regen'``, rebuild it.
        If ``'force'``, crash if not cache file found. Default ``True``.
    verbose: bool
    clean_cache_files: bool
        if ``True`` clean downloaded cache files after HDF5 are created.
    return_local_path: bool
        if ``True``, also returns the path of the local database file.
    return_partition_function: bool
        if ``True``, also returns a :py:class:`~radis.levels.partfunc.PartFuncExoMol` object.
    engine: 'vaex', 'feather'
        which memory-mapping library to use. If 'default' use the value from ~/radis.json
    output: 'pandas', 'vaex', 'jax'
        format of the output DataFrame. If ``'jax'``, returns a dictionary of
        jax arrays. If ``'vaex'``, output is a :py:class:`vaex.dataframe.DataFrameLocal`

        .. note::
            Vaex DataFrames are memory-mapped. They do not take any space in RAM
            and are extremely useful to deal with the largest databases.

    skip_optional_data : bool
        If False, fetch all fields which are marked as available in the ExoMol definition
        file. If True, load only the first 4 columns of the states file
        ("i", "E", "g", "J"). The structure of the columns above 5 depend on the
        the definitions file (*.def) and the Exomol version.
        If ``skip_optional_data=False``, two errors may occur:

            - a field is marked as present/absent in the *.def field but is
              absent/present in the *.states file (ie both files are inconsistent).
            - in the updated version of Exomol, new fields have been added in the
              states file of some species. But it has not been done for all species,
              so both structures exist. For instance, the states file of
              https://exomol.com/data/molecules/HCl/1H-35Cl/HITRAN-HCl/ follows the
              structure described in [1]_, unlike the states file of
              https://exomol.com/data/molecules/NO/14N-16O/XABC/ which follows the
              structure described in [2]_.

    Returns
    -------
    df: pd.DataFrame or vaex.dataframe.DataFrameLocal
        Line list
        A HDF5 file is also created in ``local_databases`` and referenced
        in the :ref:`RADIS config file <label_lbl_config_file>` with name
        ``databank_name``
    local_path: str
        path of local database file if ``return_local_path``

    Examples
    --------

    .. minigallery:: radis.fetch_exomol

    Notes
    -----
    if using ``load_only_wavenum_above/below`` or ``isotope``, the whole
    database is anyway downloaded and uncompressed to ``local_databases``
    fast access .HDF5 files (which will take a long time on first call). Only
    the expected wavenumber range & isotopes are returned. The .HFD5 parsing uses
    :py:func:`~radis.api.hdf5.hdf2df`

    References
    ----------

    .. [1] Tennyson, J., Yurchenko, S. N., Al-Refaie, A. F., Barton, E. J., Chubb, K. L., Coles, P. A., … Zak, E. (2016). The ExoMol database: molecular line lists for exoplanet and other hot atmospheres. https://doi.org/10.1016/j.jms.2016.05.002
    .. [2] Tennyson, J., Yurchenko, S. N., Al-Refaie, A. F., Clark, V. H. J., Chubb, K. L., Conway, E. K., … Yurchenko, O. P. (2020). The 2020 release of the ExoMol database: Molecular line lists for exoplanet and other hot atmospheres. Journal of Quantitative Spectroscopy and Radiative Transfer, 255, 107228. https://doi.org/10.1016/j.jqsrt.2020.107228

    See Also
    --------
    :py:func:`~radis.io.hitran.fetch_hitran`, :py:func:`~radis.io.hitemp.fetch_hitemp`, :py:func:`~radis.io.kurucz.fetch_kurucz`
    :py:func:`~radis.api.hdf5.hdf2df`

    """
    # TODO: implement columns= ... to load only specific columns.
    # refactor with "self._quantumNumbers" (which serves the same purpose)

    # Ensure isotope format:
    try:
        isotope = int(isotope)
    except:
        raise ValueError(
            f"In fetch_exomol, ``isotope`` must be an integer. Got `{isotope}` "
            + "Only one isotope can be queried at a time. "
        )

    full_molecule_name = get_exomol_full_isotope_name(molecule, isotope)
    known_exomol_databases, recommended_database = get_exomol_database_list(
        molecule, full_molecule_name
    )
    if verbose:
        print("\n========== Loading Exomol database [start] ==========")
    _exomol_use_hint = "Select one of them with `fetch_exomol(DATABASE_NAME)`, `SpectrumFactory.fetch_databank('exomol', exomol_database=DATABASE_NAME')`, or `calc_spectrum(..., databank=('exomol', DATABASE_NAME))` \n"
    if database is None or database == "default":
        if len(known_exomol_databases) == 1:
            database = known_exomol_databases[
                0
            ]  # TODO: if there is only one, is it not the recommended one?
        elif recommended_database:
            database = recommended_database
            if verbose > 1:
                print(
                    f"For {full_molecule_name}, the available databases are {known_exomol_databases}. {_exomol_use_hint}"
                )
        else:  # TODO: Explain here when this case occurs...
            raise KeyError(
                f"Choose one of the several databases available for {full_molecule_name} in ExoMol: {known_exomol_databases}. ({recommended_database} is recommended by the ExoMol team). {_exomol_use_hint}"
            )
    else:
        if database not in known_exomol_databases:
            raise KeyError(
                f"{database} is not of the known available ExoMol databases for {full_molecule_name}. Choose one of : {known_exomol_databases}. ({recommended_database} is recommended by the ExoMol team). {_exomol_use_hint}"
            )

    if local_databases is None:
        import radis

        local_databases = pathlib.Path(radis.config["DEFAULT_DOWNLOAD_PATH"]) / "exomol"

    local_path = (
        pathlib.Path(local_databases).expanduser()
        / molecule
        / full_molecule_name
        / database
    )

    # TODO: add deprecation if missing columns in cache file

    # Init database, download files if needed.
    mdb = MdbExomol(
        local_path,
        molecule=molecule,
        database=database,
        name=databank_name,
        local_databases=local_databases,
        nurange=[
            load_wavenum_min if load_wavenum_min is not None else 0.0,
            load_wavenum_max if load_wavenum_max is not None else np.inf,
        ],
        engine=engine,
        cache=cache,
        skip_optional_data=skip_optional_data,
        verbose=verbose,
    )

    # Get local files
    local_files = mdb.trans_file
    if not isinstance(local_files, list):
        local_files = [local_files]
    mgr = mdb.get_datafile_manager()
    local_files = [mgr.cache_file(f) for f in local_files]

    # Specific for RADIS : rename columns
    radis2exomol_columns = {
        "wav": "nu_lines",
        "airbrd": "alpha_ref",
        "El": "elower",
        "ju": "jupper",
        "jl": "jlower",
        "gp": "gupper",
        "gpp": "glower",
        "Tdpair": "n_Texp",
    }
    # get column name converting to exomol/exojax format if possible, else use the same
    if columns is not None:
        columns_exomol = [radis2exomol_columns.get(c, c) for c in columns] + [
            "Sij0",
            "jlower",
            "jupper",
        ]  # needed for broadening
    else:
        columns_exomol = None

    df = mdb.load(
        local_files,
        columns=columns_exomol,
        lower_bound=([("nu_lines", load_wavenum_min)] if load_wavenum_min else [])
        + ([("Sij0", mdb.crit)] if not np.isneginf(mdb.crit) else []),
        upper_bound=([("nu_lines", load_wavenum_max)] if load_wavenum_max else []),
        output=output,
    )

    if "jlower" not in df:
        raise KeyError(
            f"jlower not found. Maybe try to delete cache file {local_files} and restart?"
        )

    # Add broadening
    mdb.set_broadening_coef(df, output=output)

    # Specific for RADIS :
    # ... Get RADIS column names:
    exomol2radis_columns = {v: k for k, v in radis2exomol_columns.items()}
    mdb.rename_columns(df, {k: v for k, v in exomol2radis_columns.items() if k in df})

    assert "wav" in df

    # ... include isotopic abundance in linestrength :
    # Note: ExoMol treats isotopes as independent molecules ; linestrength is not
    # corrected by isotopic abundance.
    # Below, replace Linestrength with Line Intensity taking into account
    # Terrestrial isotopic abundance (to be compatible with HITRAN/HITEMP/etc. )
    from radis.db.molparam import MOLPARAMS_EXTRA_PATH, MolParams

    Ia = MolParams(extra_file_json=MOLPARAMS_EXTRA_PATH).get(
        molecule, isotope, "abundance"
    )

    if output == "jax":
        try:
            import jax.numpy as jnp
        except:
            import numpy as jnp
        df["logsij0"] += jnp.log(Ia)
    else:
        df["Sij0"] *= Ia
        mdb.rename_columns(df, {"Sij0": "int"})

    # Add Attributes of the DataFrame
    if output in ["pandas", "vaex"]:  # no attribtes in "Jax" or "Vaex" mode
        from radis.db.classes import HITRAN_MOLECULES

        attrs = {}
        if molecule in HITRAN_MOLECULES:
            attrs["id"] = get_molecule_identifier(
                molecule
            )  # HITRAN id-number (if available)
        attrs["molecule"] = molecule
        attrs["iso"] = isotope

        if output == "vaex":
            df.attrs = {}
            df.attrs = attrs
        elif output == "pandas":
            for k, v in attrs.items():
                df.attrs[k] = v
    # Return:
    out = df
    if return_local_path or return_partition_function:
        out = [out]
    if return_local_path:
        out.append(str(mdb.path))
    if return_partition_function:
        assert return_local_path
        out.append(mdb.to_partition_function_tabulator())

    if verbose:
        print("========== Loading Exomol database [end] ==========\n")
    return out



import numpy as np
import pandas as pd
import pytest

from radis.api.kuruczapi import load_pf_Barklem2016, read_kurucz
from radis.test.utils import getTestFile


@pytest.mark.fast
def test_read_kurucz():
    df = read_kurucz(
        getTestFile("gf4000.all")
    )  # from: http://kurucz.harvard.edu/atoms/4000/gf4000.all
    a = df.loc[
        5,
        [
            "orig_wavelen",
            "loggf",
            "species",
            "El",
            "jl",
            "labellower",
            "Eu",
            "ju",
            "labelupper",
            "gamRad",
            "gamSta",
            "gamvdW",
            "iso",
        ],
    ]  # .to_numpy()
    b = np.array(
        [
            504403.3610,
            -7.259,
            "40.00",
            35456.250,
            2.0,
            "d3p w3P",
            35476.070,
            3.0,
            "5s4F6s e5F",
            8.41,
            -5.27,
            -7.53,
            0,
        ],
        dtype="object",
    )
    assert (a == b).all()
    a = df.loc[
        7,
        [
            "orig_wavelen",
            "loggf",
            "species",
            "El",
            "jl",
            "labellower",
            "Eu",
            "ju",
            "labelupper",
            "gamRad",
            "gamSta",
            "gamvdW",
            "iso",
        ],
    ]  # .to_numpy()
    b = np.array(
        [
            328425.5787,
            -6.296,
            "40.00",
            21943.740,
            2.0,
            "d4 a5D",
            21974.180,
            1.0,
            "d2sp z3S",
            7.22,
            -5.99,
            -7.79,
            0,
        ],
        dtype="object",
    )
    assert (a == b).all()


@pytest.mark.fast
def test_barklem_pf():
    # Testing load_pf_Barklem2016 function
    pfTdat, pfdat = load_pf_Barklem2016()
    assert isinstance(pfTdat, pd.Series)
    assert isinstance(pfdat, pd.DataFrame)
    assert len(pfTdat) == 42
    assert len(pfdat) == 284


if __name__ == "__main__":
    test_read_kurucz()
    test_barklem_pf()



# -*- coding: utf-8 -*-
"""

::

                *(((((((
                 ((((((((((((              ,(((((
                 ((((((((((((((((/   *((((((((((*
                  ((((((((((((((((( ((((((((((((
                      (((((((( (((((((((((((
                         *
                       @@  *@@       ..  /@@
                  @@&  @@  *@@       @@  /@@  @@%
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@  (@
         ,@   @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@
         @@   @@  @@&  @@  ,.
                                    ,%&&&&&&&&&&&&&&&&&&&
          &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
           &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
             &&&&&&&&&&&&&&&&@@@@@@&@@@&&&@@@&&&&&&&&
               &&&&&&&&&&&&&&&@@@@@@&&&&&&&&&&&&&&&
                 &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
                   &&&&&&&&&&&&&&&&&&&&&&&&&&&.
                       &&&&&&&&&&&&&&&&&&&
                               .**.
                                &&&,
                                 &&

"""

import os

from .misc.config import get_config
from .misc.utils import Chdir as _chdir
from .misc.utils import getProjectRoot

# %% Config files

# @dev: refactor in progress.
# So far there are config files in ~/radis.json (for databanks), global variables
# here, and a radis/config.json file.
# Everything should be merged in a user JSON file ~/radis.json (json) overriding
# the default one.
# # molecular parameters for non-HITRAN species    should also be removed #TODO

config = get_config()
"""dict: RADIS configuration parameters

Parameters
----------
"DEBUG_MODE":False

    bool: change this at runtime with::

        import radis
        radis.config["DEBUG_MODE"] = True

    Use the :py:func:`~radis.misc.debug.printdbg` function in ``radis.misc``, typically with::

        if __debug__: printdbg(...)

    so that printdbg are removed by the Python preprocessor when running in
    optimize mode::

        python -O *.py

"ALLOW_OVERWRITE": False

    bool: Whether to allow RADIS to overwrite values of a databank entry that is already registered.
    If True, allow RADIS to download files and make changes to the local database folder even when the databank is already registered, and then re-register it with the values for the new files, overwriting the previously registered values. This could occur when e.g. running `fetch_databank` with the `cache` argument set to "regen" or the registered "paths" not including the path to the databank itself or some of the registered files missing from the disk, resulting in files being re-downloaded and the "download_date" (and possibly "paths") value being updated.
    If False, no changes are made to the local database folder when a databank is already registered, and an exception is raised in cases where re-registering would have been required, so e.g. in the example above, this setting would prevent the "download_date" from being updated and hence an exception would be raised instead.
    False by default, so RADIS only ever inserts new databank entries and doesn't modify existing ones, hence a databank entry is only ever modified by the user and responsibility is on them to ensure it is adequate for the code to run as RADIS wouldn't automatically take steps to rectify inadequate entries.

"AUTO_UPDATE_SPEC": False

    bool: experimental feature
    used to autoupdate .spec files to the latest format, by simply saving
    them again once they're loaded and fixed.
    Warning! Better have a copy of your files before that, or a way to regenerate
    them.

    Example : Add to the top of your script (once is enough!)::

        import radis
        radis.config["AUTO_UPDATE_SPEC"] = True

    You can also see: :py:func:`~radis.tools.database._update_to_latest_format`


"AUTO_UPDATE_DATABASE": False
	bool: experimental feature
	used to autoupdate .h5 files to the latest format, by simply saving
	them again once they're loaded and fixed.
	Warning! Better have a copy of your files before that, or a way to regenerate
	them.

	Example. Add to the top of your script (once is enough!)::

	    import radis
	    radis.AUTO_UPDATE_DATABASE = True

	See Also
	--------
	:py:func:`~radis.api.hdf5.hdf2df`

"MISSING_BROAD_COEF" : False

    str: Accepted values: False (default) and "air". If "air", missing boradening coefficients are replaced by those of air.

    See Also
    --------
    :py:func:`~radis.lbl.broadening._calc_broadening_HWHM`

"OLDEST_COMPATIBLE_VERSION": "0.9.1"
    str: forces to regenerate cache files that were created in a previous version

    See Also
    --------
    :py:func:`~radis.api.cache_files.load_h5_cache_file`


# "USE_CYTHON": True
#     bool: try to use Cython functions when possible

#     See more in :py:func:`radis.misc.arrays.add_at`


"GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD": 3
    float: to determine the optimal value
    of wstep using minimum FWHM value of spectrum.
    Makes sure there are enough gridpoints per line.

    See more in :  :py:meth:`radis.lbl.broadening.BroadenFactory._check_accuracy`


"GRIDPOINTS_PER_LINEWIDTH_ERROR_THRESHOLD": 1
    float: to determine the minimum feasible value
    of wstep using minimum FWHM value of spectrum.
    Makes sure there are enough gridpoints per line.

    See more in :     :py:meth:`radis.lbl.broadening.BroadenFactory._check_accuracy`

"SPARSE_WAVERANGE": True
    bool: if True, allow special optimizations to improve performances for
    large spectra that are sparse, i.e there are no lines within ``truncation``
    distance. In that case, spectral calculations can be an order of magnitude
    faster, and result in less memory use.
    This optimization may result in a small but unnecessary overhead
    when there are lines everywhere in the spectral range considered. In such
    cases, you may deactivate it by setting ``radis.config['SPARSE_WAVERANGE'] = False``
    Default ``True``

    See more in  :py:meth:`radis.lbl.broadening.BroadenFactory._apply_lineshape_LDM`

"RESAMPLING_TOLERANCE_THRESHOLD" 5e-3
    an error if raises if areas do not match by a value above this threshold,
    during resampling. See :py:meth:`~radis.spectrum.spectrum.Spectrum.resample`

    Default ``5e-3``


Notes
-----

Default values are read from the `radis/default_radis.json <https://github.com/radis/radis/blob/develop/radis/default_radis.json>`__ file.

All values are overridden at runtime by the keys in the user JSON file ``~/radis.json (json)``
(in particular, the list of databases)

See more in the :ref:`Configuration file <label_lbl_config_file>` documentation.
"""
# TODO : Refactor in progress.


# %% Version


def get_version(verbose=False, add_git_number=True):
    """Reads `__version.txt__
    <https://github.com/radis/radis/blob/master/radis/__version__.txt>`__ and
    retrieve version number. If ``add_git_number``, also appends Git commit
    number if we're on a gitted session.

    Examples
    --------

    ::

        import radis
        print(radis.get_version())
        >>> '0.9.17'
    """

    # First get version
    with open(os.path.join(getProjectRoot(), "__version__.txt")) as version_file:
        version = version_file.read().strip()

    # Now get git info
    if add_git_number:
        import subprocess
        import sys

        cd = _chdir(os.path.dirname(__file__))
        try:
            label = subprocess.check_output("git describe", stderr=subprocess.DEVNULL)
        except:
            if verbose:
                print("couldnt get git version: {0}".format(sys.exc_info()[1]))
            # probably not a git session. drop
        else:
            commit = label.decode().strip().split("-")[-1]
            version = version + "-" + commit
        finally:
            cd.__del__()

    return version


__version__ = get_version(add_git_number=False)
version = get_version(add_git_number=False)


# %% Global namespace

__all__ = [
    "config",
    "version",
    "__version__",
]

# prevent cyclic imports:
from . import api, db, io, lbl, los, misc, phys, spectrum, tools
from .api import *  # input / output common with ExoJax
from .db import *  # database of molecules
from .io import *  # input / output
from .lbl import *  # line-by-line module
from .levels import *  # rovibrational energies and partition functions
from .los import *  # line-of-sight module
from .phys import *  # conversion functions, blackbody objects
from .spectrum import *  # Spectrum object
from .test import *  # test
from .tools import *  # slit, database, line survey, etc.

__all__.extend(api.__all__)
__all__.extend(db.__all__)
__all__.extend(io.__all__)
__all__.extend(lbl.__all__)
__all__.extend(los.__all__)
__all__.extend(phys.__all__)
__all__.extend(spectrum.__all__)
__all__.extend(tools.__all__)



# -*- coding: utf-8 -*-
"""
-------------------------------------------------------------------------------
"""


import re

import pandas as pd

from radis.db.utils import getFile

# Isotope definitions in HITRAN
isotope_name_dict = {
    (1, 1): "H2(16O)",
    (1, 2): "H2(18O)",
    (1, 3): "H2(17O)",
    (1, 4): "HD(16O)",
    (1, 5): "HD(18O)",
    (1, 6): "HD(17O)",
    (1, 7): "D2(16O)",
    (2, 1): "(12C)(16O)2",
    (2, 2): "(13C)(16O)2",
    (2, 3): "(16O)(12C)(18O)",
    (2, 4): "(16O)(12C)(17O)",
    (2, 5): "(16O)(13C)(18O)",
    (2, 6): "(16O)(13C)(17O)",
    (2, 7): "(12C)(18O)2",
    (2, 8): "(17O)(12C)(18O)",
    (2, 9): "(12C)(17O)2",
    (2, 10): "(13C)(18O)2",
    (2, 11): "(18O)(13C)(17O)",
    (2, 12): "(13C)(17O)2",
    (3, 1): "(16O)3",
    (3, 2): "(16O)(16O)(18O)",
    (3, 3): "(16O)(18O)(16O)",
    (3, 4): "(16O)(16O)(17O)",
    (3, 5): "(16O)(17O)(16O)",
    (4, 1): "(14N)2(16O)",
    (4, 2): "(14N)(15N)(16O)",
    (4, 3): "(15N)(14N)(16O)",
    (4, 4): "(14N)2(18O)",
    (4, 5): "(14N)2(17O)",
    (5, 1): "(12C)(16O)",
    (5, 2): "(13C)(16O)",
    (5, 3): "(12C)(18O)",
    (5, 4): "(12C)(17O)",
    (5, 5): "(13C)(18O)",
    (5, 6): "(13C)(17O)",
    (6, 1): "(12C)H4",
    (6, 2): "(13C)H4",
    (6, 3): "(12C)H3D",
    (6, 4): "(13C)H3D",
    (7, 1): "(16O)2",
    (7, 2): "(16O)(18O)",
    (7, 3): "(16O)(17O)",
    (8, 1): "(14N)(16O)",
    (8, 2): "(15N)(16O)",
    (8, 3): "(14N)(18O)",
    (9, 1): "(32S)(16O)2",
    (9, 2): "(34S)(16O)2",
    (10, 1): "(14N)(16O)2",
    (11, 1): "(14N)H3",
    (11, 2): "(15N)H3",
    (12, 1): "H(14N)(16O)3",
    (12, 2): "H(15N)(16O)3",
    (13, 1): "(16O)H",
    (13, 2): "(18O)H",
    (13, 3): "(16O)D",
    (14, 1): "H(19F)",
    (14, 2): "D(19F)",
    (15, 1): "H(35Cl)",
    (15, 2): "H(37Cl)",
    (15, 3): "D(35Cl)",
    (15, 4): "D(37Cl)",
    (16, 1): "H(79Br)",
    (16, 2): "H(81Br)",
    (16, 3): "D(79Br)",
    (16, 4): "D(81Br)",
    (17, 1): "H(127I)",
    (17, 2): "D(127I)",
    (18, 1): "(35Cl)(16O)",
    (18, 2): "(37Cl)(16O)",
    (19, 1): "(16O)(12C)(32S)",
    (19, 2): "(16O)(12C)(34S)",
    (19, 3): "(16O)(13C)(32S)",
    (19, 4): "(16O)(12C)(33S)",
    (19, 5): "(18O)(12C)(32S)",
    (20, 1): "H2(12C)(16O)",
    (20, 2): "H2(13C)(16O)",
    (20, 3): "H2(12C)(18O)",
    (21, 1): "H(16O)(35Cl)",
    (21, 2): "H(16O)(37Cl)",
    (22, 1): "(14N)2",
    (22, 2): "(14N)(15N)",
    (23, 1): "H(12C)(14N)",
    (23, 2): "H(13C)(14N)",
    (23, 3): "H(12C)(15N)",
    (24, 1): "(12C)H3(35Cl)",
    (24, 2): "(12C)H3(37Cl)",
    (25, 1): "H2(16O)2",
    (26, 1): "(12C)2H2",
    (26, 2): "(12C)(13C)H2",
    (26, 3): "(12C)2HD",
    (27, 1): "(12C)2H6",
    (27, 2): "(12C)H3(13C)H3",
    (28, 1): "(31P)H3",
    (29, 1): "(12C)(16O)(19F)2",
    (29, 2): "(13C)(16O)(19F)2",
    (30, 1): "(32S)(19F)6",
    (31, 1): "H2(32S)",
    (31, 2): "H2(34S)",
    (31, 3): "H2(33S)",
    (32, 1): "H(12C)(16O)(16O)H",
    (33, 1): "H(16O)2",
    (34, 1): "(16O)",
    (35, 1): "(35Cl)(16O)(14N)(16O)2",
    (35, 2): "(37Cl)(16O)(14N)(16O)2",
    (36, 1): "(14N)(16O)+",
    (37, 1): "H(16O)(79Br)",
    (37, 2): "H(16O)(81Br)",
    (38, 1): "(12C)2H4",
    (38, 2): "(12C)H2(13C)H2",
    (39, 1): "(12C)H3(16O)H",
    (40, 1): "(12C)H3(79Br)",
    (40, 2): "(12C)H3(81Br)",
    (41, 1): "(12C)H3(12C)(14N)",
    (42, 1): "(12C)(19F)4",
    (43, 1): "(12C)4H2",
    (44, 1): "H(12C)3(14N)",
    (45, 1): "H2",
    (45, 2): "HD",
    (46, 1): "(12C)(32S)",
    (46, 2): "(12C)(34S)",
    (46, 3): "(13C)(32S)",
    (46, 4): "(12C)(33S)",
    (47, 1): "(32S)(16O)3",
    (48, 1): "(12C)2(14N)2",
    (49, 1): "(12C)(16O)(35Cl)2",
    (49, 2): "(12C)(16O)(35Cl)(37Cl)",
    (49, 2): "(12C)(16O)(35Cl)(37Cl)",
    (50, 1): "(32S)(16O)",
    (50, 2): "(34S)(16O)",
    (50, 3): "(32S)(18O)",
    (51, 1): "(12C)H3(19F)",
    (52, 1): "(74Ge)H4",
    (52, 2): "(72Ge)H4",
    (52, 3): "(70Ge)H4",
    (52, 4): "(73Ge)H4",
    (52, 5): "(76Ge)H4",
    (53, 1): "(12C)(32S)2",
    (53, 2): "(32S)(12C)(34S)",
    (53, 3): "(32S)(12C)(33S)",
    (53, 4): "(13C)(32S)2",
    (54, 1): "(12C)H3(127I)",
    (55, 1): "(14N)(19F)3",
}
# TODO : add test in test_molparams comparing dictionary above to
# latest values from HAPI; to make sure we never have any problem
# even if HITRAN eventually changes the conventions and labels.


#%%
# Read Extra parameters taken from molparams_extra.json
# TODO : Refactor : have either .json either .txt, but not both

from json import JSONDecodeError
from os.path import exists, join

import hjson

from radis.misc.utils import getProjectRoot

# %% Functions to parse radis/config.json

MOLPARAMS_EXTRA_PATH = join(getProjectRoot(), "db", "molparams_extra.json")
assert exists(MOLPARAMS_EXTRA_PATH)


def get_extra_molparams(path=MOLPARAMS_EXTRA_PATH):
    """Read the extra molecular parameters in
    :py:attr:`~radis.db.molparam.MOLPARAMS_EXTRA_PATH` to complement the
    HITRAN molecular parameters of molparam.txt"""

    with open(path) as f:
        try:
            molparams_dict = hjson.load(f)
            # read with Hjson to allow comments in MOLPARAMS_EXTRA_PATH
        except JSONDecodeError as err:
            raise JSONDecodeError(
                "Error reading '{0}' (line {2} col {3}): \n{1}".format(
                    path, err.msg, err.lineno, err.colno
                ),
                err.doc,
                err.pos,
            ) from err

    return molparams_dict


#%%


class MolParams(object):
    """Easy access to molecular parameters taken from HITRAN `molparam.txt`.

    .. note::
        Starting from Radis 0.13.1 MolParams can also read extra parameters
        from a .json file

    Parameters
    ----------
    file: str
        if None the one in RADIS is taken. See https://github.com/radis/radis/blob/master/radis/db/molparam.txt
    extra_file_json: str
        if not ``None``, read extra molecular parameters from a Json file, for
        molecules not implemented in HITRAN  `molparam.txt`.
        See :py:attr:`~radis.db.molparam.MOLPARAMS_EXTRA_PATH`

    Examples
    --------
    Get earth abundance of CO2, isotope 1::

        from radis.db.molparam import MolParams
        molpar = MolParams()
        molpar.get(2, 1, 'abundance')         # 2 for CO2, 1 for isotope 1

    .. minigallery:: radis.db.molparam.MolParams
    :add-heading:

    Note
    ----
    Isotope number was derived manually assuming the isonames were ordered in the database
    The isotope name (ex: CO2 626) is kept for comparison if ever needed

    References
    ----------
    http://hitran.org/media/molparam.txt
    """

    __slots__ = ["df", "terrestrial_abundances", "extra_molparams"]

    def __init__(
        self,
        file=None,
        terrestrial_abundances=True,
        extra_file_json=MOLPARAMS_EXTRA_PATH,
    ):
        if file is None:
            file = getFile("molparam.txt")

        df = pd.read_csv(file, comment="#", sep="\s+")
        df = df.set_index(["id", "iso"])

        df["isotope_name_exomol"] = _add_exomol_name(df)

        self.df = df
        self.terrestrial_abundances = terrestrial_abundances
        self.extra_molparams = {}

        if extra_file_json is not None:
            self.extra_molparams.update(get_extra_molparams(extra_file_json))

        # ------
        try:  # Add hints (Python >3.6 only)
            self.get.__annotations__["key"] = list(df.keys())
        except AttributeError:
            pass  # old Python version

    def _get_molecule_list(self):
        """Returns list of [molecule_id, molecule_name] in loaded molparam dataframe

        Examples
        --------
        ::

            [1,  'H2O',  2,  'CO2',  3,  'O3', 4, 'N2O',  etc.  ]
        """
        from radis.db.classes import get_molecule

        return [
            item
            for m in self.df.reset_index(inplace=False)["id"].unique()
            for item in [m, get_molecule(m)]
        ]

    def get(self, molecule, isotope, key):
        r"""Get attribute of molecule, isotope.

        Parameters
        ----------
        molecule: str or int
            molecule name or molecule id
        iso: int
            isotope number # (sorted by terrestrial abundance)
        key: ``'abundance'``, ``'mol_mass'``, ``'isotope_name'``, ``'isotope_name_exomol'``
            parameter

        Examples
        --------
        Get explicit name of an isotope::

            from radis.db.molparam import MolParams
            mp = MolParams()
            print(mp.get("CO2", 1, "isotope_name"))   # >> (12C)(16O)2
            print(mp.get("CO2", 2, "isotope_name"))   # >> (13C)(16O)2


        .. minigallery:: radis.db.molparam.MolParams.get
            :add-heading:

        """
        if key == "mol_mass":  # backward compatibility >=0.13.1
            key = "molar_mass"

        if key not in self.df.columns:
            raise ValueError(
                f"Unknown molecular parameter : {key}. Use one of {self.df.columns}"
            )

        M = molecule
        try:
            try:
                float(M)
            except ValueError:
                # M is not a number (i.e. not the HITRAN molecule id)
                from radis.db.classes import get_molecule_identifier

                M = get_molecule_identifier(M)
        except NotImplementedError:
            # Molecule not in HITRAN . Try to read from extra parameters
            if key in self.extra_molparams and M in self.extra_molparams[key]:
                try:
                    return self.extra_molparams[key][M][str(isotope)]
                except KeyError:

                    raise NotImplementedError(
                        f"{key} for isotope `{str(isotope)}` for molecule {M} not supported"
                        + "\nChoose one of{0}".format(
                            list(self.extra_molparams[key][M].keys())
                        )
                        + f"\nYou can manually add it in the {MOLPARAMS_EXTRA_PATH} file"
                        + "{'molparams':{'molar_mass':{'molecule':{'ISOTOPE':...}}}}; or in the `SpectrumFactory.molpar.extra_molparams[key][molecule][isotope]` dictionary. Please also report on GitHub so we can update !"
                    )
            else:
                raise NotImplementedError(
                    f"{key} for molecule '{M}' not supported"
                    + "\nChoose one of{0}".format(self._get_molecule_list())
                    + "\nor `\n{0}".format(
                        list(self.extra_molparams.get(key, {}).keys())
                    )
                    + f"\nYou can manually add it in the {MOLPARAMS_EXTRA_PATH} file"
                    + "{'molparams':{'molar_mass':{'molecule':{'ISOTOPE':...}}}}; or in the `SpectrumFactory.molpar.extra_molparams[key][molecule][isotope]` dictionary. Please also report on GitHub so we can update !"
                )

        else:
            out = self.df.loc[(M, isotope), key]
            assert len(out) == 1
            return out.iloc[0]


def _add_exomol_name(df):
    r"""Convert HITRAN isotope name from :py:data:`~radis.db.molparam.isotope_name_dict`
    in ExoMol format
    """

    df["isotope_name"] = df.index.map(isotope_name_dict)

    # add exomol full name:
    def replace_parenthesis(s):
        s = str(s)
        # add parenthesis for hydrogen : Eg. H2 -> (1H)2,   H->(1H)
        s = re.sub(r"H([1-9])*", r"(1H)\1", s)
        # add parenthesis for deuterium : Eg. D2 -> (2H)2,   D->(2H)
        s = re.sub(r"D([1-9])*", r"(2H)\1", s)

        # Cut elements and replace parenthesis with - :
        elements = re.findall(r"\(.*?\)[1-9]*", s)
        elements = [e.replace(r"(", "").replace(")", "") for e in elements]
        s = "-".join(elements)

        return s

    return df.isotope_name.apply(replace_parenthesis)


if __name__ == "__main__":
    from radis.test.io.test_exomol import test_exomol_parsing_functions

    test_exomol_parsing_functions()

    from radis.test.db.test_molparams import test_molparams

    test_molparams()



# -*- coding: utf-8 -*-
"""
Test conversion functions
"""

import astropy.units as u

from radis.phys.units_astropy import convert_and_strip_units


def test_convert():

    # Dimensioned input: should be converted
    assert convert_and_strip_units(4500 * u.nm, u.um) == 4.5
    assert convert_and_strip_units(4.5 * u.um, u.nm) == 4500  # uses round(10)
    assert convert_and_strip_units(2000 * (1 / u.cm), 1 / u.m) == 200000

    # Convert using the spectral equivalencies (astropy.units.equivalencies)
    assert convert_and_strip_units(5 * u.um, 1 / u.cm) == 2000
    assert convert_and_strip_units(60 * u.THz, 1 / u.cm) == 2001.3845711889

    # Covert using temperature equivalencies
    assert convert_and_strip_units(0 * u.deg_C, u.K) == 273.15
    assert convert_and_strip_units(300 * u.K, u.imperial.deg_F) == 80.33

    # Undimensioned input : shouldnt change
    assert convert_and_strip_units(4500, u.um) == 4500

    # None type : should return None
    assert convert_and_strip_units(None) is None
    assert convert_and_strip_units(None, u.nm) is None


if __name__ == "__main__":
    test_convert()



# -*- coding: utf-8 -*-
"""Physical constants.

Note : more units available in :
- scipy.constants
- sympy.physics.units


-------------------------------------------------------------------------------
"""

# %% SI units

eV = 1.602176634e-19  # J
"""float: electron volt (J)

https://physics.nist.gov/cgi-bin/cuu/Value?e
"""

eV2wn = 8065.54  # 1[eV]=8065.54[cm^-1] # from: https://github.com/HajimeKawahara/exojax/blob/78466cef0170ee1a2768b6a6f7b7c911d715c1bd/src/exojax/utils/constants.py#L35

h = 6.62607015e-34
"""float: Planck constant (m2.kg.s-1 = J.s)

https://physics.nist.gov/cgi-bin/cuu/Value?h|search_for=planck"""

e_CGS = 4.80320450e-10
"""[esu]=[dyn^0.5*cm] #elementary charge"""

m_e_CGS = 9.10938356e-28
"""[g] !electron mass"""

k_b = 1.3806490e-23
"""float: Boltzmann constant (m2.kg.s-2.K-1)

https://physics.nist.gov/cgi-bin/cuu/Value?k|search_for=boltzmann
"""

c = 2.99792458e8
"""float: light velocity (m/s)

https://physics.nist.gov/cgi-bin/cuu/Value?c
"""

a_0 = 5.29177e-11
"""float: First Bohr radius (m)"""

m_e = 9.1093837015e-31
"""float: Electron mass (kg)

https://physics.nist.gov/cgi-bin/cuu/Value?me|search_for=electron+mass"""

Ry = 13.605693122994
"""float: Rydberg unit of energy (eV)

https://physics.nist.gov/cgi-bin/cuu/Value?rydhcev|search_for=rydberg"""

Av = 6.02214076e23
"""float: Avogadro constant (molecules / mol)

https://physics.nist.gov/cgi-bin/cuu/Value?na|search_for=avogadro"""
Na = Av  # just an alias

eps_0 = 8.8541878128e-12
"""float: Vacuum permittivity (Farads / m)

https://physics.nist.gov/cgi-bin/cuu/Value?ep0|search_for=vacuum+permittivity"""

hc_k = h * c / k_b * 100  #
"""float:  cm to K conversion (~ 1.44 K/cm), used in Boltzmann factors"""


# %% HITRAN (CGS) units

k_b_CGS = 1.380648813e-16
"""float: Boltzmann constant (erg / K)"""

c_CGS = 2.99792458e10
"""float: Light velocity (cm / s)"""

h_CGS = 6.626196e-27
"""float: Planck constant (erg * s)"""



# -*- coding: utf-8 -*-
"""Physical constants and conversion.
"""

from .blackbody import planck, planck_wn, sPlanck
from .convert import (
    J2K,
    K2J,
    J2cm,
    J2eV,
    K2eV,
    atm2bar,
    atm2torr,
    bar2atm,
    bar2torr,
    cm2eV,
    cm2hz,
    cm2J,
    cm2nm,
    dcm2dnm,
    dnm2dcm,
    eV2cm,
    eV2J,
    eV2K,
    eV2nm,
    hz2cm,
    nm2cm,
    nm2eV,
    torr2atm,
    torr2bar,
)
from .units import (
    conv2,
    convert_emi2cm,
    convert_emi2nm,
    convert_rad2cm,
    convert_rad2nm,
    is_homogeneous,
)
from .units_astropy import convert_and_strip_units

__all__ = ["sPlanck", "planck", "planck_wn"]
